using System;
using System.Net;
using System.Text;
using System.Web;
using System.Security.Cryptography;
using System.Security.Cryptography.X509Certificates;

using com.chinaeinv.einv.order.dto;
using Newtonsoft.Json;

namespace com.chinaeinv.einv.order
{
    /**
    * 订单接口代理实现
    */
    public class OrderFacadeAgent : OrderFacade {
        private string appCode, keyStorePath, password, apiUrl;

        /**
        * 构造方法
        * @param appCode 由电子发票平台分配的appCode
        * @param keyStorePath 密钥库文件路径
        * @param alias 密钥库别名
        * @param password 密钥库密码
        * @param apiUrl 接口URL
        */
        public OrderFacadeAgent(string appCode, string keyStorePath, string password, string apiUrl) {
            this.appCode = appCode;
            this.keyStorePath = keyStorePath;
            this.password = password;
            this.apiUrl = apiUrl;
        }

        /**
        * (异步)开具蓝字发票
        */
        public Result KpAsync(Order order) {
            return Exec<Result, Order>(order, "chinaeinv.api.order.v11.kp_async");
        }

        /**
        * (异步)开具红字发票
        */
        public Result ChAsync(Order order) {
            return Exec<Result, Order>(order, "chinaeinv.api.order.v11.ch_async");
        }

        /**
        * (异步)对开票失败的订单重新开票
        */
        public Result RetryKpAsync(OrderBase orderBase) {
            return Exec<Result, OrderBase>(orderBase, "chinaeinv.api.order.v11.retryKp_async");
        }

        /**
        * (同步)取消订单
        */
        public Result Cancel(OrderBase orderBase) {
            return Exec<Result, OrderBase>(orderBase, "chinaeinv.api.order.v11.cancel");
        }

        /**
        * (同步)根据订单编号查询发票
        */
        public InvoiceResult CxByOrderNo(OrderBase orderBase) {
            return Exec<InvoiceResult, OrderBase>(orderBase, "chinaeinv.api.order.v11.cx.orderNo");
        }

        /**
        * 调用接口
        * @param order 请求数据
        * @return 响应数据
        * @throws Exception
        */
        private T Exec<T,O>(O order, string cmdName) {
            // 将Order对象转为json
            string requestJson = JsonConvert.SerializeObject(order, Formatting.None, new JsonSerializerSettings { NullValueHandling = NullValueHandling.Ignore });
            Console.WriteLine("方法名称: " + cmdName + ", 请求报文: " + requestJson);

            // 生成签名
            string sign = GetSign(requestJson);
            Console.WriteLine("方法名称: " + cmdName + ", 签名: " + sign);

            // 调用接口
            string responseJson;
            using (var client = new WebClient())
            {
                client.Encoding = Encoding.UTF8;
                client.QueryString.Add("appCode", appCode);
                client.QueryString.Add("cmdName", cmdName);
                client.QueryString.Add("sign", UrlEncode(UrlEncode(sign, Encoding.UTF8), Encoding.UTF8)); // C#中需要进行两次UrlEncode，不然发送请求时加号会变空格，导致验签失败。
                responseJson = client.UploadString(apiUrl, requestJson);
            }
            Console.WriteLine("方法名称: " + cmdName + ", 响应报文: " + responseJson);

            // 将响应json转为对象
            return JsonConvert.DeserializeObject<T>(responseJson);
        }

        /**
        * 数字签名
        */
        private string GetSign(string kpParams)
        {
            X509Certificate2 x = new X509Certificate2(keyStorePath, password, X509KeyStorageFlags.Exportable);
            RSACryptoServiceProvider rsa = x.PrivateKey as RSACryptoServiceProvider;
            byte[] data = Encoding.UTF8.GetBytes(kpParams);
            RSACryptoServiceProvider privateKey1 = new RSACryptoServiceProvider();
            privateKey1.ImportParameters(rsa.ExportParameters(true));
            byte[] sign = privateKey1.SignData(data, "MD5");
            return Convert.ToBase64String(sign);
        }

        private string UrlEncode(string temp, Encoding encoding)
        {
            StringBuilder stringBuilder = new StringBuilder();
            for (int i = 0; i < temp.Length; i++)
            {
                string t = temp[i].ToString();
                string k = HttpUtility.UrlEncode(t, encoding);
                if (t == k)
                {
                    stringBuilder.Append(t);
                }
                else
                {
                    stringBuilder.Append(k.ToUpper());
                }
            }
            return stringBuilder.ToString();
        }
    }
}